__version__ = '0.0.2'
from .abba import XABBA, fastXABBA, fastXABBA_len, fastXABBA_inc

from .comp import compress
from .agg import aggregate 
from .inverse import *

from .llmabba import LLMABBA

__name__ =  'llmabba'
